import os, io, pickle

import unittest as ut
import unittest.mock as m

from context import *


class TestCurveProcessing(ut.TestCase):
    def setUp(self):
        '''
        x = [ 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 19 18 17
        16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0  1  2  3  4  5  6  7
        8]

        y = [ 0  1  2  3  4  5  6  7  8  9 10  9  8  7  6  5  6  7  8  9 10 11 12 13
        14 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0  0  1  2  3  4  5  6
        7  8]

        gc = [  0  -1  -2  -3  -4  -5  -6  -7  -8  -9 -10 -11 -12 -13 -14 -15 -16 -17
        -18 -19 -20 -19 -18 -17 -16 -15 -14 -13 -12 -11 -10  -9  -8  -7  -6  -5
        -6  -7  -8  -9 -10  -9  -8  -7  -6  -5  -4  -3  -2  -1]
        '''

        x = [i for i in range(20)] + [i for i in range(20, 0, -1)] + [i for i in range(0, 10)]
        y = [i for i in range(10)] + [i for i in range(10, 5, -1)] + [i for i in range(5, 15)] + [i for i in range(15, -1, -1)] + [i for i in range(10)]
        gc = [i for i in range(0, -20, -1)] + [i for i in range(-20, -5)] + [i for i in range(-5, -10, -1)] + [i for i in range(-10, 1)]

        self.x = np.asarray(x)
        self.y = np.asarray(y)
        self.gc = np.asarray(gc)


    def test_process_curve_1(self):
        res_x_min = sorted(CurveProcessing.process_curve(self.x, 'min', 2))
        res_x_max = sorted(CurveProcessing.process_curve(self.x, 'max', 2))
        res_y_min = sorted(CurveProcessing.process_curve(self.y, 'min', 2))
        res_y_max = sorted(CurveProcessing.process_curve(self.y, 'max', 2))
        res_gc_min = sorted(CurveProcessing.process_curve(self.gc, 'min', 2))
        res_gc_max = sorted(CurveProcessing.process_curve(self.gc, 'max', 2))

        exp_x_min = [Peak(0, 50, 2), Peak(40, 50, 2)]
        exp_x_max = []
        exp_y_min = []
        exp_y_max = []
        exp_gc_min = []
        exp_gc_max = []

        self.assertEqual(exp_x_min, res_x_min)
        # self.assertEqual(exp_x_max, res_x_max)
        # self.assertEqual(exp_y_min, res_y_min)
        # self.assertEqual(exp_y_max, res_y_max)
        # self.assertEqual(exp_gc_min, res_gc_min)
        # self.assertEqual(exp_gc_max, res_gc_max)
    

    def test_curve_combinations_1(self):
        ...


    def test_detect_peaks_1(self):
        ...
    

    def test_filter_peaks_1(self):
        ...



if __name__ == '__main__':
    ut.main()


"""Module for handing biodata files and whatnot."""

import os
import csv
import pickle
import warnings
from typing import Optional
from io import TextIOWrapper
from urllib.error import HTTPError, URLError

from Bio import SeqIO, Entrez

from orcapy.Peak import Peak

warnings.simplefilter('always', UserWarning)


def fetch_file(accession: str, email: str, rettype: str, api_key: Optional[str] = None) -> TextIOWrapper:
    """Downloads the given file_type of the given accession in temporary memory"""
    Entrez.email = email
    if api_key is not None:
        Entrez.api_key = api_key
    try:
        return Entrez.efetch(db="nuccore", id=accession, rettype=rettype, retmode="text")
    except HTTPError as BadRequest:
        if api_key is not None:
            api_message_1 = f' with API_key \'{api_key}\''
            api_message_2 = ' and if your API_key is correctly typed and linked to the given email'
        else:
            api_message_1, api_message_2 = '', ''
        BadRequest.add_note(
            f'Unable to fetch accession: \'{accession}\' using \'{email}\'{api_message_1}. Please check if the accession is of an existing chromosomal sequence{api_message_2}.')
        raise
    except URLError as NoConnection:
        NoConnection.add_note(
            'You are fetching a file from the NCBI servers. Please make sure you have an internet connection to do so.')
        raise


def parse_SeqRecord(record: SeqIO.SeqRecord, genes_of_interest: list[str]) -> dict:
    """
    Extracts the sequence and positions of the genes of interest from a SeqRecord.
    Will throw a warning if the file was already annotated with an oriC. This oriC
    is not used in the computation.
    """
    accession, version = tuple(record.id.split('.'))
    seq_dict = {
        'accession': accession,
        'version': int(version),
        'seq': str(record.seq),  # I only use the sequence to loop over once. strings are much faster for this.
        'seq_len': len(record.seq),
        'gene_locations': [],
        'NCBI_oriC': []
    }

    for feature in record.features:
        # is this feature a coding sequence and a gene and is its name something we are looking for?
        if feature.type == 'CDS' and 'gene' in feature.qualifiers and feature.qualifiers['gene'][0] in genes_of_interest:
            gene_loc = Peak.from_calc_middle(int(feature.location.start), int(feature.location.end), len(record.seq), 0)
            seq_dict['gene_locations'].append((feature.qualifiers['gene'][0], gene_loc))
        # just in case this SeqRecord has an annotated oriC!
        if feature.type == 'rep_origin':
            oriC = Peak.from_calc_middle(int(feature.location.start), int(feature.location.end), len(record.seq), 0)
            seq_dict['NCBI_oriC'].append(('oriC', oriC))
            warnings.warn(
                f"{accession}.{version}: found an annotated oriC in the provided file. Saved in 'NCBI_oriC' attribute of the ORCA object.",
                UserWarning
            )
    if len(seq_dict["NCBI_oriC"]) == 0:
        seq_dict['NCBI_oriC'] = None
    return seq_dict


def save_gbk(accession: str, email: str, output_folder: str, api_key: Optional[str] = None):
    '''
    Download the GenBank file of a given accession and save it into the output_folder.
    Name of the file will be: `{accession}_{version}.gbk`.
    Raises a `FileExistsError` if a file with the generated name already exists in the output folder.
    If version is not provided in the accession, then the function downloads the latest version.
    '''
    with fetch_file(accession, email, rettype="gbwithparts", api_key=api_key) as fh:
        # Quick search for the version of the accession that was downloaded.
        acc_v = "."
        header = ""
        for line in fh:
            # Underlying stream is not seekable, so cannot reset back to top of file
            # -> need to keep track of what has been read
            header += line
            # Always appears around the top of the file.
            if "VERSION" in line:
                acc_v = [x for x in line.strip("\n").split(" ")][-1]
                break
        acc, version = tuple(acc_v.split("."))

        # Check if a file with the same name already exists
        if os.path.exists(os.path.join(output_folder, acc + '_' + version + '.gbk')):
            fh.close()
            raise FileExistsError(f'\'{acc}_{version}.gbk\' already exists in: {output_folder}')

        # Save contents to path
        file_path = os.path.join(output_folder, acc + '_' + version + '.gbk')
        with open(file_path, 'w') as oh:
            oh.write(header)
            oh.write(fh.read())
            oh.close()
        fh.close()


def save_pkl(accession: str, email: str, output_folder: str, api_key: Optional[str] = None):
    '''
    Download the GenBank file of a given accession, parses it with Biopython into a SeqRecord, and save it into the output_folder.
    Name of the file will be: `{accession}_{version}.pkl`.
    Raises a `FileExistsError` if a file with the generated name already exists in the output folder.
    If version is not provided in the accession, then the function downloads the latest version.
    '''
    with fetch_file(accession, email, rettype="gbwithparts", api_key=api_key) as fh:

        # Parse gbk file
        seq_rec = SeqIO.read(fh, 'gb')
        acc, version = seq_rec.id.split('.')

        # Check if a file with the same name already exists
        if os.path.exists(os.path.join(output_folder, acc + '_' + version + '.pkl')):
            fh.close()
            raise FileExistsError(f'\'{acc}_{version}.pkl\' already exists in: {output_folder}')

        # Save contents to path
        file_path = os.path.join(output_folder, acc + '_' + version + '.pkl')
        with open(file_path, 'wb') as oh:
            pickle.dump(seq_rec, oh)
            oh.close()
        fh.close()


def merge_csvs(file_folder: str, merged_csv: str, fieldnames: list[str], length: int = -1, headers: bool = False):
    '''
    Merge multiple csvs into one csv.
    Arguments:
    - file_folder : path to folder with csvs that have to be merged
    - merged_csv  : name of the merged csv
    - fieldnames  : list of fieldnames for the merged_csv
    - length      : amount of rows in each single csv. -1, if the length of each single
                    csv is not known or not the same.
    - headers     : if the single csvs have headers or not
    '''
    file_list = os.listdir(file_folder)
    with open(merged_csv, 'w', newline='') as fh_out:
        writer = csv.writer(fh_out)
        writer.writerow(fieldnames)
        for file in file_list:
            with open(os.path.join(file_folder, file), 'r') as fh_in:
                reader = csv.reader(fh_in)
                for i, row in enumerate(reader):
                    if headers and i == 0:
                        pass
                    elif i == length:
                        break
                    else:
                        writer.writerow(row)



import unittest as ut

from context import *

# remove later
class TestPeak(ut.TestCase):
    def setUp(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        / c    a b           f          h                                                     g    d        e
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        '''
        self.a = Peak(60, 1000, 60)
        self.b = Peak(80, 1000, 60)
        self.c = Peak(10, 1000, 60)
        self.d = Peak(900, 1000, 60)
        self.e = Peak(990, 1000, 60)

        self.f = Peak(200, 1000, 60)
        self.g = Peak(850, 1000, 60)
        self.h = Peak(310, 1000, 60)

        self.lst1 = [self.a, self.b, self.c, self.d, self.e]
        self.lst2 = [self.f, self.g, self.h]


    def test_init_1(self):
        x = Peak(0, 0, 0)
        self.assertIsInstance(x, Peak)


    def test_init_2(self):
        x = Peak(8, 10, 8)
        self.assertIsInstance(x, Peak)


    def test_from_calc_middle(self):
        x = Peak(5, 10, 2)
        y = Peak.from_calc_middle(3, 7, 10, 2)
        self.assertEqual(x, y)


    def test_from_edges(self):
        x = Peak(5, 10, 2)
        y = Peak.from_edges(4, 6, 10)
        self.assertEqual(x, y)


    def test_calc_dist_points_1(self):
        res = Peak.calc_dist_points(self.a.middle, self.b.middle, self.a.seq_len)
        self.assertEqual(res, 20)

    
    def test_calc_dist_points_2(self):
        res = Peak.calc_dist_points(self.a.middle, self.d.middle, self.a.seq_len)
        self.assertEqual(res, 160)


    def test_calc_dist_1(self):
        res = self.a.calc_dist(self.b)
        self.assertEqual(res, 20)

    
    def test_calc_dist_2(self):
        res = self.a.calc_dist(self.d.middle)
        self.assertEqual(res, 160)


    def test_get_peaks_to_merge(self):
        res = Peak.get_peaks_to_merge(self.lst1)
        exp = [(self.a, self.b), (self.c, self.e)]
        self.assertEqual(res, exp)


    def test_match_peaks(self):
        x = [self.a, self.b, self.c]
        y = [self.c, self.d, self.e]

        res = sorted(Peak.match_peaks(x, y))
        exp = sorted([(self.c, self.c), (self.a, self.c), (self.c, self.e)])
        self.assertEqual(res, exp)


    def test_get_adjacency_matrix_1(self):
        res = Peak.get_adjacency_matrix(self.lst1)
        exp = np.array([
            [  0,  20,  50, 160,  70],
            [ 20,   0,  70, 180,  90],
            [ 50,  70,   0, 110,  20], 
            [160, 180, 110,   0,  90],
            [ 70,  90,  20,  90,   0]
        ])
        self.assertTrue(np.array_equal(res, exp))


    def test_get_adjacency_matrix_2(self):
        res = Peak.get_adjacency_matrix(self.lst1, self.lst2)
        exp = np.array([
            #  f,   g,   h
            [140, 210, 250], # a
            [120, 230, 230], # b
            [190, 160, 300], # c
            [300,  50, 410], # d
            [210, 140, 320]  # e
        ])
        self.assertTrue(np.array_equal(res, exp))


    def test_get_adjacency_matrix_3(self):
        lst1 = [x.middle for x in self.lst1]
        lst2 = [x.middle for x in self.lst2]
        res = Peak.get_adjacency_matrix(lst1, lst2, seq_len=1000)
        exp = np.array([
            #  f,   g,   h
            [140, 210, 250], # a
            [120, 230, 230], # b
            [190, 160, 300], # c
            [300,  50, 410], # d
            [210, 140, 320]  # e
        ])
        self.assertTrue(np.array_equal(res, exp))
    

    def test_get_adjacency_matrix_4(self):
        lst2 = [x.middle for x in self.lst2]
        with self.assertRaisesRegex(ValueError, "Elements is `peaks_a` are not the same type as `peaks_b`."):
            Peak.get_adjacency_matrix(self.lst1, lst2)


    def test_get_adjacency_matrix_5(self):
        lst2 = [self.f, self.g.middle, self.h]
        with self.assertRaisesRegex(ValueError, "Not all elements in `peaks_b` are of the same type."):
            Peak.get_adjacency_matrix(self.lst1, lst2)


    def test_get_adjacency_matrix_5(self):
        lst1 = [self.f, self.g.middle, self.h]
        with self.assertRaisesRegex(ValueError, "Not all elements in `peaks_a` are of the same type."):
            Peak.get_adjacency_matrix(lst1)


    def test_get_adjacency_matrix_7(self):
        lst1 = [x.middle for x in self.lst2]
        with self.assertRaisesRegex(ValueError, "Provided list of integers, but did not provide `seq_len`."):
            Peak.get_adjacency_matrix(lst1)


    def test_get_adjacency_matrix_8(self):
        lst2 = [self.f, self.g.middle, self.h]
        with self.assertRaisesRegex(ValueError, "Not all elements in `peaks_b` are of the same type."):
            Peak.get_adjacency_matrix(self.lst1, lst2)


    def test_get_connected_components_idx_1(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        / i   i i                                         jj   j     j                                      i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /________                                         ____________                                      __
        '''

        i1 = Peak(  0, 1000, 60)
        i2 = Peak( 40, 1000, 60)
        i3 = Peak( 60, 1000, 60)
        i4 = Peak(990, 1000, 60)

        j1 = Peak(490, 1000, 60)
        j2 = Peak(500, 1000, 60)
        j3 = Peak(540, 1000, 60)
        j4 = Peak(600, 1000, 60)

        peaks = [i1, i2, i3, i4, j1, j2, j3, j4]
        mat = Peak.get_adjacency_matrix(peaks)

        res = sorted(Peak.get_connected_components_idx([i for i in range(len(peaks))], mat, 120))
        exp = sorted([[0, 1, 2, 3], [4, 5, 6, 7]])
        self.assertEqual(res, exp)


    def test_get_connected_components_idx_2(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        /i  i  i  i  i  i  i  i  i  i  i  i  i  i  i  i        i  i  i  i  i  i  i  i  i  i  i  i  i  i  i  i 
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /______________________________________________        _______________________________________________
        '''
        lst_peaks = [Peak(i, 1000, 60) for i in range(0, 1000, 20)]
        indexes = [i for i in range(len(lst_peaks))]

        mat = Peak.get_adjacency_matrix(lst_peaks)

        res = Peak.get_connected_components_idx(indexes, mat, 50)
        exp = [indexes]
        self.assertEqual(res, exp)


    def test_get_connected_components_idx_3(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        /      i  i  i  i  i  i  i  i  i  i  i  i  i  i        i  i  i  i  i  i  i  i  i  i  i  i  i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /      ________________________________________        _____________________________________
        '''
        lst_peaks_1 = [Peak(i, 1000, 60) for i in range(60, 450, 20)]
        lst_peaks_2 = [Peak(i, 1000, 60) for i in range(540, 910, 20)]
        indexes = [i for i in range(len(lst_peaks_1 + lst_peaks_2))]

        mat = Peak.get_adjacency_matrix(lst_peaks_1 + lst_peaks_2)

        res = Peak.get_connected_components_idx(indexes, mat, 50)
        exp = [[i for i in range(len(lst_peaks_1))], [i+len(lst_peaks_1) for i in range(len(lst_peaks_2))]]
        self.assertEqual(res, exp)


    def test_select_connected_groups_2(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        / i   i i                                         jj   j     j                                      i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /________                                         ______     _                                      __
        '''
        i1 = Peak(  0, 1000, 60)
        i2 = Peak( 40, 1000, 60)
        i3 = Peak( 60, 1000, 60)
        i4 = Peak(990, 1000, 60)

        j1 = Peak(490, 1000, 60)
        j2 = Peak(500, 1000, 60)
        j3 = Peak(540, 1000, 60)
        j4 = Peak(600, 1000, 60)

        peaks = [i1, i2, i3, i4, j1, j2, j3, j4]
        mat = Peak.get_adjacency_matrix(peaks)
        res = sorted(Peak.select_connected_groups(peaks, mat, 50))
        exp = sorted([[i1, i2, i3, i4], [j1, j2, j3], [j4]])
        self.assertEqual(res, exp)


    def test_select_connected_groups_3(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        / i   i i  i   i i                                jj   j   j                                        i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /__   ___  _   ___                                ______   _                                        __
        '''
        i1 = Peak(  0, 1000, 60)
        i2 = Peak( 40, 1000, 60)
        i3 = Peak( 60, 1000, 60)
        i4 = Peak(990, 1000, 60)
        i5 = Peak(100, 1000, 60)
        i6 = Peak(140, 1000, 60)
        i7 = Peak(160, 1000, 60)

        j1 = Peak(490, 1000, 60)
        j2 = Peak(500, 1000, 60)
        j3 = Peak(540, 1000, 60)
        j4 = Peak(600, 1000, 60)

        peaks = [i1, i2, i3, i4, i5, i6, i7, j1, j2, j3, j4]
        mat = Peak.get_adjacency_matrix(peaks)
        # print("-------------------------------------------------")
        res = sorted(Peak.select_connected_groups(peaks, mat, 50))
        # print([[str(i) for i in x] for x in res])
        exp = sorted([[i1, i4], [i2, i3], [i5], [i6, i7], [j1, j2, j3], [j4]])
        # print("-------------------------------------------------")
        self.assertEqual(res, exp)


    def test_select_connected_groups_4(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        / i   i i                                jjjjjjjjjj jjjjjjjjjj                                      i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /________                                __________ __________                                      __
        '''
        i1 = Peak(  0, 1000, 60)
        i2 = Peak( 40, 1000, 60)
        i3 = Peak( 60, 1000, 60)
        i4 = Peak(990, 1000, 60)

        jlst1 = [Peak(400+i, 1000, 60) for i in range(0, 100, 10)]
        jlst2 = [Peak(510+i, 1000, 60) for i in range(0, 90, 10)]

        peaks = [i1, i2, i3, i4] + jlst1 + jlst2
        mat = Peak.get_adjacency_matrix(peaks)

        res = sorted(Peak.select_connected_groups(peaks, mat, 50))
        # print([[str(i) for i in x] for x in res])
        exp = sorted([[i1, i2, i3, i4], jlst1, jlst2])
        self.assertEqual(res, exp)


    def test_select_connected_groups_5(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10
        / i   i i                           j j  jjj  jjj j jjj j jj j   j                                  i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /________                           ___  ___  ___ _ ___ _ __ _   _                                  __
        '''
        i1 = Peak(  0, 1000, 60)
        i2 = Peak( 40, 1000, 60)
        i3 = Peak( 60, 1000, 60)
        i4 = Peak(990, 1000, 60)

        js = [350, 370, 400, 410, 420, 450, 460, 470, 490, 510, 520, 530, 550, 570, 580, 600, 640]
        j = [Peak(i, 1000, 60) for i in js]

        peaks = [i1, i2, i3, i4] + j
        mat = Peak.get_adjacency_matrix(peaks)

        res = sorted(Peak.select_connected_groups(peaks, mat, 50))
        # print([[str(i) for i in x] for x in res])
        exp = sorted([[i1, i2, i3, i4], j[0:2], j[2:5], j[5:8], j[8:9], j[9:12], j[12:13], j[13:15], j[15:16], j[16:]])
        # print([[str(i) for i in x] for x in exp])
        self.assertEqual(res, exp)


    def test_select_connected_groups_6(self):
        '''
        /0         1         2         3         4         5         6         7         8         9         10        1         2         3         4         5         6         7         8         9         20
        / i   i i                           j j  jjj  jjj j jjj j jj j   j                                  i i   i i                           j j  jjj  jjj j jjj j jj j   j                                  i
        /|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|---------|
        /________                           ___________________________  _                                  _________                           ___________________________  _                                  __
        '''
        i1 = Peak(   0, 2000, 60)
        i2 = Peak(  40, 2000, 60)
        i3 = Peak(  60, 2000, 60)
        i4 = Peak( 990, 2000, 60)

        i5 = Peak(1000, 2000, 60)
        i6 = Peak(1040, 2000, 60)
        i7 = Peak(1060, 2000, 60)
        i8 = Peak(1990, 2000, 60)

        js = [350, 370, 400, 410, 420, 450, 460, 470, 490, 510, 520, 530, 550, 570, 580, 600, 640]
        j1 = [Peak(i, 2000, 60) for i in js]
        j2 = [Peak(i+1000, 2000, 60) for i in js]

        peaks = [i1, i2, i3, i4, i5, i6, i7, i8] + j1 + j2
        mat = Peak.get_adjacency_matrix(peaks)

        # twice as long so 5% = 100
        # would be perfect if all j's were together, but this is a tuning issue, not a testing issue
        # print('======================================================')
        res = sorted(Peak.select_connected_groups(peaks, mat, 100))
        # print([[str(i) for i in x] for x in res])
        # print('======================================================')
        exp = sorted([[i1, i2, i3, i8], [i4, i5, i6, i7], j1[:-1], [j1[-1]], j2[:-1], [j2[-1]]])
        self.assertEqual(res, exp)


    def test_calc_middle_1(self):
        self.assertEqual(Peak.calc_middle(self.a, self.b), 70)


    def test_calc_middle_2(self):
        self.assertEqual(Peak.calc_middle(self.a, self.e), 26)


    def test_calc_middle_3(self):
        self.assertEqual(Peak.calc_middle(self.a, self.g), 955)


    def test_calc_middle_4(self):
        with self.assertRaisesRegex(ValueError, "No curve size found in any parameters. This method is for circular DNA only."):
            Peak.calc_middle(5, 3)


    def test_intersecting_windows_1(self):
        self.assertTrue(self.a.intersecting_windows(self.a))
        self.assertTrue(self.a.intersecting_windows(self.b))
        self.assertTrue(self.a.intersecting_windows(self.c))
        self.assertFalse(self.a.intersecting_windows(self.d))
        self.assertFalse(self.a.intersecting_windows(self.e))
        self.assertFalse(self.b.intersecting_windows(self.e))
        self.assertTrue(self.c.intersecting_windows(self.e))


    def test_intersecting_windows_2(self):
        with self.assertRaisesRegex(ValueError, "other is a <class 'float'>, must be a Peak object"):
            self.a.intersecting_windows(3.5)


    def test_contains_point(self):
        self.assertTrue(self.a.contains_point(40))
        self.assertTrue(self.a.contains_point(self.b))
        self.assertTrue(self.b.contains_point(self.a))
        self.assertFalse(self.a.contains_point(self.d))


    def test_contains(self):
        self.assertTrue(40 in self.a)
        self.assertTrue(self.b in self.a)
        self.assertTrue(self.a in self.b)
        self.assertFalse(self.d in self.a)


    def test_hash(self):
        b = Peak(60, 1000, 60)
        self.assertEqual(hash(self.a), hash(b))
        self.assertNotEqual(hash(self.a), hash(self.b))


    def test_eq(self):
        self.assertEqual(self.a, self.a)
        self.assertEqual(self.a, Peak(60, 1000, 60))
        self.assertNotEqual(self.a, self.b)
        self.assertNotEqual(self.a, 3)


    def test_repr(self):
        s = 'Peak(middle=60, window_size=60, seq_len=1000)'
        self.assertEqual(self.a.__repr__(), s)


    def test_str(self):
        self.assertEqual(str(self.a), "60")


    def test_lt(self):
        self.assertTrue(self.a < self.b)
        self.assertFalse(self.a < self.a)


    def test_gt(self):
        self.assertTrue(self.b > self.a)
        self.assertFalse(self.b > self.b)


    def test_add(self):
        self.assertEqual(Peak(70, 1000, 60), self.a + self.c)
        self.assertEqual(70, self.a + 10)


    def test_sub(self):
        self.assertEqual(Peak(50, 1000, 60), self.a - self.c)
        self.assertEqual(50, self.a - 10)


    def test_radd(self):
        self.assertEqual(160, 100 + self.a)


    def test_rsub(self):
        self.assertEqual(40, 100 - self.a)
    

    def test_neg(self):
        self.assertEqual(Peak(-self.a.middle, self.a.seq_len, self.a.window_size), -self.a)


    def test_float(self):
        self.assertEqual(float(self.a), 60.0)


if __name__ == '__main__':
    ut.main()


# Libraries
import os
import pickle
import re
import warnings
from itertools import combinations, product
from typing import Generator, Optional

import numpy as np
from Bio import Seq, SeqIO
from sklearn.base import is_classifier

# Self-made modules
from orcapy import BioFile, CurveProcessing, Plotter
from orcapy.Peak import Peak

warnings.simplefilter('always', UserWarning)


class ORCA:
    """
    ORCA: Origin of Chromosomal Replication Assessment. Class for predicting the location of the origin of replication (ORC/oriC) of circular bacterial DNA.
    For further instruction, consult the README on [GitHub](https://github.com/ZoyavanMeel/ORCA/).

    Accepted **kwargs:
    ------------------
    - `dnaa_boxes`:
        - If None, will use the [consensus DnaA-box](https://doi.org/10.1093/bib/bbn031): `TTAT(A|T|C|G)CACA`.
        Else, provide a list of 9 base strings. See the `get_dnaa_boxes_with_mismatches` function for some more examples of dnaA-boxes.
        Example input: `['AAAAAAAAA', 'TTTTTTTTT']`.
    - `max_mismatches`:
        - Maximum allowed mismatches allowed in a dnaa_box for it still to be read as such. Recommended max: 2.
    - `genes_of_interest`:
        - List of gene names to consider as 'oriC-proximal' and use for helping estimate the location of the oriC. This parameter is case insensitive.
    - `max_point_spread`:
        - Maximum distance between points in a group can have when looking for connected groups across the disparity curves.
        Default is 5 % of the total chromosome length.
    - `windows`:
        - The windows around peaks of skew curves to consider. Defaults are 1, 3, and 5 % of the total chromosome length. ORCA checks each of the given windows.
    - `model`:
        - A fitted scikit-learn classifier. Recommended to use the one provided on [GitHub](https://github.com/ZoyavanMeel/ORCA/).
    """

    def __init__(self):
        """
        Empty constructor. Use: `from_gbk()`, `from_pkl()`, `from_accession()` or `from_string()` for instantiating ORCA objects.
        """
        pass

    def __make(self, **kwargs):
        """Private: used to check the validity of user-given keyword arguments and set them as instance variables."""
        user_args = ORCA.setup_variables(**kwargs)
        for key, value in user_args.items():
            setattr(self, key, value)

    @classmethod
    def from_string(cls, sequence: str, gene_locations: list[tuple[str, int, int]] = None, accession: str = None, **kwargs) -> "ORCA":
        """
        Minimal constructor: provide a string to analyse and whatever else you want.

        Parameters:
        ----------
        - `sequence`        : String representation of the DNA circular sequence to analyse.
        - `gene_locations`  : Optional, list of tuples in the following format: `(gene_name, start_position, end_position)`.
        - `accession`       : Optional, used for naming, else: 'Custom'.
        - `**kwargs`        : Check the ORCA class documentation for valid keyword arguments.

        Returns:
        --------
        - `ORCA` : an ORCA object.

        --------------------------
        Example:
        --------

        >>> # Bio.Seq objects are also fine, they just take longer to loop over.
        >>> DNA = "ATCGATCGATCGATACGATGTGCTAGCTACTGATCGATCGACAGACTGCTAGCGATCCTCGA"
        >>> gene_locs = [("dnaA", 10, 30), ("dnaN", 40, 50)]
        >>> # To instantiate the ORCA object call:
        >>> orca = ORCA.from_string(DNA, gene_locs)
        >>> # To find oriCs:
        >>> orca.find_oriCs()
        """
        orca = cls()
        orca.__make(**kwargs)

        orca.seq = sequence
        orca.seq_len = len(sequence)
        orca.version = 0

        orca.accession = "Custom" if accession is None else accession
        orca.gene_locations = [] if gene_locations is None \
            else [(gene[0], Peak.from_calc_middle(gene[1], gene[2], len(sequence), 0)) for gene in gene_locations]

        return orca

    @classmethod
    def from_gbk(cls, path: str, **kwargs) -> "ORCA":
        """
        Instatiates an ORCA object from a GenBank (gbk) file.

        Parameters:
        ----------
        - `path`     : string of path to the GenBank file to analyse.
        - `**kwargs` : Check the ORCA class documentation for valid keyword arguments.

        Returns:
        --------
        - `ORCA` : an ORCA object with properly loaded parameters.

        --------------------------
        Example:
        -------

        >>> # To instantiate the ORCA object call:
        >>> orca = ORCA.from_gbk("path/to/file.gbk")
        >>> # To find oriCs:
        >>> orca.find_oriCs()
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"{path} does not exist.")

        orca = cls()
        orca.__make(**kwargs)

        file_args = BioFile.parse_SeqRecord(SeqIO.read(path, 'gb'), orca.genes_of_interest)
        for key, value in file_args.items():
            setattr(orca, key, value)
        return orca

    @classmethod
    def from_pkl(cls, path: str, **kwargs) -> "ORCA":
        """
        Instatiates an ORCA object from a pickled Biopython SeqRecord.

        Parameters:
        -----------
        - `path`     : string of path to the pickled SeqRecord of a GenBank file.
        - `**kwargs` : Check the ORCA class documentation for valid keyword arguments.

        Returns:
        --------
        - `ORCA` : an ORCA object with properly loaded parameters.

        Raises:
        -------
        - `FileNotFoundError` : in case the file is not found at `path` 
        - `EOFError`          : in case of faulty/empty Pickle file.
        - `ValueError`        : in case the pickled object is not a SeqRecord.

        --------------------------
        Example:
        --------

        >>> # To instantiate the ORCA object call:
        >>> orca = ORCA.from_pkl("path/to/file.pkl")
        >>> # To find oriCs:
        >>> orca.find_oriCs()
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"{path} does not exist.")

        orca = cls()
        orca.__make(**kwargs)

        with open(path, "rb") as fh:
            try:
                record = pickle.load(fh)
            except EOFError as eof:
                eof.add_note(
                    f"Something is wrong with your Pickle file. It is probably empty. ORCA was unable to load: \'{path}\'.")
                fh.close()
                raise
            fh.close()
            if not isinstance(record, SeqIO.SeqRecord):
                raise ValueError(f"Pickled object must be a SeqRecord object, but was a \'{type(record)}\' instead.")
            file_args = BioFile.parse_SeqRecord(record, orca.genes_of_interest)
            for key, value in file_args.items():
                setattr(orca, key, value)
            return orca

    @classmethod
    def from_accession(cls, accession: str, email: str, api_key: str = None, **kwargs) -> "ORCA":
        """
        Instatiates an ORCA object from a given accession number.

        Parameters:
        -----------
        - `accession` : accession to analyse. ORCA will fetch and parse the required file.
        - `email`     : your email address. Always tell NCBI who you are.
        - `api_key`   : Optional. Only necessary if you are doing more than 10 requests per second.
        - `**kwargs`  : Check the ORCA class documentation for valid keyword arguments.

        Returns:
        --------
        - `ORCA` : an ORCA object with properly loaded parameters.

        --------------------------
        Example:
        --------

        >>> # To instantiate the ORCA object call:
        >>> orca = ORCA.from_accession("NC_example.3", "your@email.com")
        >>> # To find oriCs:
        >>> orca.find_oriCs()
        """
        orca = cls()
        orca.__make(**kwargs)

        if email is None:
            raise ValueError(f"""
                NCBI requires you to specify an email address to make use of their services.
                \tCreate an NCBI account at: https://www.ncbi.nlm.nih.gov/
                \tGenerate an API_key at: https://www.ncbi.nlm.nih.gov/account/settings/ (optional)
            """)
        with BioFile.fetch_file(accession, email, rettype='gbwithparts', api_key=api_key) as handle:
            file_args = BioFile.parse_SeqRecord(SeqIO.read(handle, 'gb'), orca.genes_of_interest)
            handle.close()
            for key, value in file_args.items():
                setattr(orca, key, value)
            return orca

    @staticmethod
    def setup_variables(**kwargs) -> dict:
        """
        Handle the user input parameters and throw errors and warnings if they are not valid. 
        Check the ORCA class documentation for valid keyword arguments.
        """

        defaults = {
            'dnaa_boxes': ['TTATACACA', 'TTATTCACA', 'TTATCCACA', 'TTATGCACA'],
            'genes_of_interest': ['dnaA', 'dnaN'],
            'windows': [0.01, 0.03, 0.05],
            'max_point_spread': 0.05,
            'max_mismatches': 0,
            'model': None,
        }

        # Update defaults with provided values
        args = {key: kwargs[key] if key in kwargs else defaults[key] for key in defaults}

        # Simple type enforcing. Overriding default parameters is okay, but provide empty lists if you do.
        assert args['model'] is None or is_classifier(args['model']), 'model must be a scikit-learn classifier.'
        assert isinstance(args['windows'], list), 'windows must be a list of integers.'
        assert isinstance(args['dnaa_boxes'], list), 'dnaa_boxes must be a list of strings.'
        assert isinstance(args['max_mismatches'], int), 'max_mismatches must be an integer.'
        assert isinstance(args['max_point_spread'], float), 'max_point_spread must be an integer.'
        assert isinstance(args['genes_of_interest'], list), 'genes_of_interest must be a list of strings.'

        # Setup all possible dnaA-box regions
        args['input_dnaa_boxes'] = args['dnaa_boxes'].copy()
        args['dnaa_boxes'] = ORCA.get_dnaa_boxes_with_mismatches(args['input_dnaa_boxes'], args['max_mismatches'])

        return args

    @staticmethod
    def get_dnaa_boxes_with_mismatches(box_list: list, max_mismatches: int = 2) -> list:
        """
        Standard parameters: Get all unique dnaa-box 9-mers and reverse complements that allow for 0, 1, or 2 mismatches.
        Sources as comments in function. Check source code for more suggested consensus boxes.

        Parameters:
        - `box_list`       : list with strings of dnaa-boxes to use. Make sure all given boxes are 9 bases long.
        - `max_mismatches` : maximum allowed mismatches in dnaa-box that is still seen as an accepted dnaa-box.
            E.g. 2 allows all kmers that have 0, 1 or 2 mismatches with the dnaa-box.\n
        Return:
            `dnaa_boxes`   : set of 9-mers matching the given parameters.

        ------------------------------------------------------------
        Useful acticles about DnaA(-boxes):
        - https://doi.org/10.1101/cshperspect.a012922
        - https://doi.org/10.1093/nar/gkr832
        - https://doi.org/10.3389/fmicb.2018.00319
        - https://doi.org/10.1046/j.1365-2958.1996.6481362.x.

        ------------------------------------------------------------

        |                                Sequence                                      |                    DOI                       |  Year  |                     Notes                         |
        | ---------------------------------------------------------------------------- | -------------------------------------------- | ------ | ------------------------------------------------- |
        | 'TTAT(A\|C)CA(A\|C)A'                                                        | https://doi.org/10.1016/0092-8674(84)90284-8 |  1984  | The first consensus                               |
        | 'TGTG(G\|T)ATAAC'                                                            | https://doi.org/10.1016/0022-2836(85)90299-2 |  1985  | Matsui-box                                        |
        | 'TTAT(A\|T\|C\|G)CACA'                                                       | https://doi.org/10.1093/dnares/dsm017        |  2007  | In (roughly) all eubacteria, not just B. subtilis |
        | 'TTATCCACA', 'TTTTCCACA', 'TTATCAACA', 'TCATTCACA', 'TTATACACA', 'TTATCCAAA' | https://doi.org/10.1093/emboj/16.21.6574     |  1997  | Affinity study                                    |
        | '(T\|C)(T\|C)(A\|T\|C)T(A\|C)C(A\|G)(A\|C\|T)(A\|C)'                         | https://doi.org/10.1007/BF00273584           |  1991  | Only in E. coli K12. Do not use.                  |
        """

        if max_mismatches == 0:
            return box_list

        def generate_mismatched_strings(string: str, mismatches: int = 2) -> Generator:
            string_len = len(string)
            letters = 'ACGT'

            for indices in combinations(range(string_len), mismatches):
                for replacements in product(letters, repeat=mismatches):
                    skip = False
                    for i, a in zip(indices, replacements):
                        if string[i] == a:
                            skip = True
                    if skip:
                        continue

                    keys = dict(zip(indices, replacements))
                    yield ''.join([string[i] if i not in indices else keys[i] for i in range(string_len)])

        # Get all dnaa-boxes as strings
        boxes = box_list.copy()
        for box in box_list:
            if re.search(r'[^ATCG]', box) is not None:
                raise ValueError(
                    f'\n\tInput string: \'{box}\' contains forbidden characters. Only use the four nucleotide bases: A, T, C, and G.')
            # Also check reverse complement for box on other strand
            boxes.append(str(Seq.Seq(box).reverse_complement()))

        # Get all unique strings while allowing for max. 2 mismatches.
        mismatch_boxes = list(set(boxes))
        for box in boxes:
            for i in range(abs(max_mismatches)):
                mismatch_boxes.extend(generate_mismatched_strings(box, i+1))
        return mismatch_boxes

    def calculate_disparity_curves(self) -> None:
        '''
        Z-curve and GC-skew calculation and k-mer indexing. In one function so only one iteration of the sequence is necessary.\n
        Sets:
            `x`, `y`, `z`, `gc` : 1D-np.arrays of the three Z-curve components and 1D-np.array of the GC-skew
            `dnaa_dict`         : Dictionary of middle indexes of dnaa-boxes in `seq`
        '''
        k = len(self.dnaa_boxes[0])
        x, y, z, gc = [], [], [], []
        a, c, t, g = 0, 0, 0, 0

        raw_dict = {}

        for i in range(self.seq_len):
            base = self.seq[i]
            if base == "A":
                a += 1
            elif base == "C":
                c += 1
            elif base == "G":
                g += 1
            elif base == "T":
                t += 1

            gc.append(g - c)
            x.append((a + g) - (c + t))  # Purine vs Pyrimidine
            y.append((a + c) - (g + t))  # Amino vs Keto
            z.append((a + t) - (c + g))  # Weak vs Strong Hydrogen Bonds

            kmer = self.seq[i:i+k] if i <= self.seq_len - k else self.seq[i:] + self.seq[:k-(self.seq_len-i)]
            if kmer in self.dnaa_boxes:
                mid = i+k//2+1 if i+k//2+1 <= self.seq_len else i+k//2+1 - self.seq_len
                try:
                    raw_dict[kmer].append(mid)
                except KeyError:
                    raw_dict[kmer] = [mid]

        (self.x, self.y, self.z, self.gc, self.dnaa_dict) = (
            np.asarray(x), np.asarray(y), np.asarray(z), np.asarray(gc), raw_dict)

    def analyse_disparity_curves(self) -> list[Peak]:
        '''
        Analyse the three disparity curves related to finding the oriC of the sequence.
        The Z-curve can only be analysed after calling `calc_disparities()`.
        '''
        peaks = []
        for fraction in self.windows:
            window_size = int(self.seq_len * fraction)
            peaks_x = CurveProcessing.process_curve(self.x, 'min', window_size=window_size)
            peaks_y = CurveProcessing.process_curve(self.y, 'max', window_size=window_size)
            peaks_gc = CurveProcessing.process_curve(self.gc, 'min', window_size=window_size)
            peaks.extend([j for i in CurveProcessing.curve_combinations(
                (peaks_x, peaks_y, peaks_gc), self.seq_len) for j in i])
        return peaks

    def calculate_Z_scores(self, peaks: list[Peak]) -> tuple[list[Peak], list[int]]:
        '''Finding connected components in undirected graph with a depth-first search to merge Z-curve oriCs'''

        adj_mat = Peak.get_adjacency_matrix(peaks)
        connected_groups = Peak.select_connected_groups(peaks, adj_mat, int(self.seq_len*self.max_point_spread))
        connected_groups.sort(key=lambda x: len(x), reverse=True)

        total_pot_oriCs = len([y for x in connected_groups for y in x])
        oriCs, Z_scores = [], []
        window_size = int(self.seq_len*self.windows[-1])

        for group in connected_groups:
            group.sort()
            for i in range(len(group)):
                if (group[-1] - group[i]).middle >= (self.seq_len-1)/2:
                    group[i].middle += self.seq_len-1
            avg_val = sum(group)//len(group)
            if avg_val > self.seq_len-1:
                avg_val -= self.seq_len-1
            oriCs.append(Peak(avg_val, self.seq_len, window_size))
            Z_scores.append(len(group)/total_pot_oriCs)
        return oriCs, Z_scores

    def calculate_D_scores(self, peaks: list[Peak]) -> list[Peak]:
        '''Process the location of dnaa_boxes and rank potential oriCs based on most surrounding dnaa_boxes.'''
        if len(self.dnaa_dict.keys()) == 0:
            return [0] * len(peaks)
        contains_boxes = []
        all_pos = [pos for pos_list in self.dnaa_dict.values() for pos in pos_list]
        for oriC in peaks:
            count = 0
            for pos in all_pos:
                if oriC.contains_point(pos):
                    count += 1
            contains_boxes.append(count)
        return [x/sum(contains_boxes) if sum(contains_boxes) != 0 else 0 for x in contains_boxes]

    def calculate_G_scores(self, peaks: list[Peak]) -> list[int]:
        '''Process the location of the genes of interest and rank the potential oriCs based on how close they are to these genes'''
        if len(self.gene_locations) == 0:
            return [0] * len(peaks)
        gene_peaks = [name_loc_tuple[1] for name_loc_tuple in self.gene_locations]
        matrix = Peak.get_adjacency_matrix(peaks, gene_peaks)
        if np.min(matrix) == np.max(matrix):
            return [1] * matrix.shape[1]
        norm_mat = (matrix - np.min(matrix)) / (np.max(matrix) - np.min(matrix))
        return [1 - x for x in np.mean(norm_mat, axis=1)]

    def __check_run_properly(self, attr: str, msg: str) -> None:
        """Used internally to check if the ORCA object has the proper attributes it needs at certain stages."""
        try:
            getattr(self, attr)
        except AttributeError as NotRunYet:
            NotRunYet.add_note(msg)
            raise

    def pretty_print_results(self) -> None:
        """Print the main ORCA results in a pretty way."""
        self.__check_run_properly("oriC_middles", "Cannot print results.")

        spacer_1 = len('oriC_middles')+1
        spacer_2 = len(str(max(self.oriC_middles)))+2

        print(f'{"Accession":<{spacer_1}}: {self.accession}')

        attrs = ['predictions', 'Z_scores', 'G_scores', 'D_scores', 'oriC_middles']
        for attr in attrs:
            if isinstance(getattr(self, attr)[0], float):
                results = ''.join(f'{round(x, spacer_2-4): >{spacer_2}}' for x in getattr(self, attr))
            else:
                results = ''.join(f'{str(x): >{spacer_2}}' for x in getattr(self, attr))
            print(f'{attr:<{spacer_1}}:{results}')
        print("The best-scoring potential oriC was found at:", self.oriC_middles[self.best_oriC_idx], "bp.")

    def plot_oriC_curves(self, plot_path: str) -> None:
        """
        Plot curves relevant to ORCA's analysis as well as all found oriCs.
        To be called after calling `find_oriCs` on an ORCA object.
        See Plotter for more plotting options.

        --------------------------------
        Example:

        >>> from orcapy import ORCA, Plotter
        >>> orca = ORCA.from_accession("NC_example.3", "your@email.com")
        >>> orca.find_oriCs()
        >>> orca.plot_oriC_curves("path/to/plot.png")

        The same as calling:

        >>> from orcapy import ORCA, Plotter
        >>> orca = ORCA.from_accession('NC_000913', 'example@email.com')
        >>> orca.find_oriCs()
        >>> Plotter.plot_curves(path="path/to/plot.png", curves=[orca.x, orca.y, orca.gc], peaks=orca.oriC_middles, labels=['$x_n$', '$y_n$', '$GC_n$'], name=orca.accession)

        Also the same as calling:

        >>> from orcapy import ORCA, Plotter
        >>> orca = ORCA.from_accession("NC_example.3", "your@email.com")
        >>> orca.find_oriCs(plot_path="path/to/plot.png")
        """
        self.__check_run_properly("oriC_middles", "Cannot plot curves.")
        return Plotter.plot_curves(
            path=plot_path,
            curves=[self.x, self.y, self.gc],
            labels=['$x_n$', '$y_n$', '$GC_n$'],
            peaks=self.oriC_middles,
            name=self.accession
        )

    def _set_best_pot_oriC_idx(self) -> None:
        """
        Set the best performing potential oriC based on the model's prediction as well as a tie-breaking method.

        Initially the potential oriC is chosen in which the model has the most "confidence", i.e. which one does
        the model think is the most likely to be the True origin. This is determined by the highest probability of
        "correctness" as calculated by the `predict_proba` function of the classifier.

        In the case of a tie, we look at the highest score of the most important feature (Z-score) between the
        tied potential oriCs. If this is also results in some form of tie, we check the next one, etc.
        """

        # set matrix columns to put them in order of feature importance: [prediction, Z-score, D-score, G-score]
        if len(self.predictions) == 0 or self.predictions[0] is None:
            mat = np.asarray([self.Z_scores, self.D_scores, self.G_scores]).T
        else:
            mat = np.asarray([self.predictions, self.Z_scores, self.D_scores, self.G_scores]).T
        max_idx_options = [j for j in range(mat.shape[0])]

        for i in range(mat.shape[1]):
            max_idx_in_curr_col = np.where(mat[max_idx_options, i] == np.max(mat[max_idx_options, i]))[0].tolist()
            max_idx_options = [max_idx_options[j] for j in max_idx_in_curr_col]
            if len(max_idx_options) == 1:
                break
        self.best_oriC_idx = max_idx_options[0]

    def find_oriCs(self, show_info: bool = False, plot_path: Optional[str] = None) -> None:
        '''
        Locates potential oriCs on circular bacterial chromosomes based on Z-curve and GC-skew analysis, dnaA box analysis, and dnaA/dnaN gene locations.
        Three default window_sizes are used: 1, 3 and 5 % of the total genome length. See the README-file in the [GitHub repository](https://github.com/ZoyavanMeel/ORCA/)
        for more information or consult ORCA.pdf for more extensive results and analyses of ORCA.

        Parameters:
        -----------
        - `show_info`:
            - If True, prints info of ALL found oriCs. Good and bad.

        - `plot_path`:
            - If not `None`, makes a plot of ALL found oriCs. Good and bad. Should not be used for analysis -> Make a separate plot for the best oriCs.
            - Will be saved to the specified file path.

        Results
        -------
        Adds new instance variables to the ORCA object based on the analysis.

        All attributes after calling `find_oriCs`:
        - `accession`                           : `str`                     Accesssion number.
        - `version`                             : `int`                     Accession version number.
        - `input_dnaa_boxes`                    : `list[str]`               List of user-given dnaA-boxes to look for.
        - `dnaa_boxes`                          : `list[str]`               List of all possible accepted dnaA-boxes based on the given dnaA-boxes and the allowed `max_mismatches`.
        - `dnaa_dict`                           : `dict[str : list[int]]`   Dictionary of dnaA-box indexes on the DNA sequence.
        - `max_mismatches`                      : `int`                     Maximum allowed mismatches in the proposed dnaA-boxes.
        - `genes_of_interest`                   : `list[str]`               List of user-given genes that relate to the location of oriCs.
        - `gene_locations`                      : `list[tuple[str, Peak]]`  List of locations of the `genes_of_interest`.
        - `NCBI_oriC`                           : `list[tuple[str, Peak]]`  Only if the file/accesssion analysed had an annotated 'rep_origin', else `None`
        - `windows`                             : `list[int]`               Windows surrounding potential oriCs. Used for filtering, distance approximation, and generalisations
        - `max_point_spread`                    : `float`                   Maximum distance between points in a component as a fraction of the length of the sequence
        - `model`                               : `sklearn.classifier`      Scikit-learn classifier used for predicting the significance of the found oriCs.
        - `seq`                                 : `str`                     String representation of the DNA sequence
        - `seq_len`                             : `int`                     Length of the DNA sequence
        - `x`, `y`, `z`, `gc`                   : `numpy.ndarray`           Three components of the Z-curve as well as the GC-skew curve
        - `oriCs`                               :`list[Peak]`               List of Peak object of the predicted origin locations. Consult its scores and prediction
                                                                            for analysis of its validity.
        - `oriC_middles`                        : `list[int]`               List of only the middle posistion point of each predicted oriC in the same order as `oriCs`.
        - `Z_scores`, `G_scores`, `D_scores`    : `list[float]`             List of scores based on the analysis of the Z-curve, gene locations, and dnaA-box regions.
                                                                            Consult the paper for more information. In the same order as `oriCs`.
        - `predictions`                         : `list[float]`             List of decision values decided by the given `model`. In the same order as `oriCs`.

        --------------------------
        Example:
        --------

        >>> # To instantiate the ORCA object call:
        >>> orca = ORCA.from_accession("NC_example.3", "your@email.com")
        >>> # To find oriCs:
        >>> orca.find_oriCs()
        >>> print(orca.accession)
        NC_example.3
        '''

        self.__check_run_properly("accession", "This object has not been instantiated properly.")

        # Step 1: Calculate disparity curves + finding dnaA-box regions
        self.calculate_disparity_curves()

        # Step 2: Finding potential oriCs based on disparity curves
        peaks_of_interest = self.analyse_disparity_curves()
        oriCs, Z_scores = self.calculate_Z_scores(peaks_of_interest)

        # Step 3: DnaA-box analysis
        if len(self.dnaa_dict.keys()) == 0:
            warnings.warn(f'''Accession: {self.accession + "." + str(self.version)}.
                No DnaA-boxes were found: {self.input_dnaa_boxes}
                Allowed for {self.max_mismatches} mismatches.
                This will affect the validity of the prediction!
            ''', UserWarning)
        D_scores = self.calculate_D_scores(oriCs)

        # Step 4: Gene-location analysis
        if len(self.gene_locations) == 0:
            warnings.warn(f'''Accession: {self.accession + "." + str(self.version)}.
                None of the genes of interest were found in the given data: {self.genes_of_interest}
                This will affect the validity of the prediction!
            ''', UserWarning)
        G_scores = self.calculate_G_scores(oriCs)

        # Step 5: Machine Learning model decision function.
        decisions = [None for i in range(len(oriCs))]
        if self.model is not None:
            decisions = self.model.predict_proba(np.asarray(
                [Z_scores, G_scores, D_scores]).T)[:, 1].tolist()
        oriC_middles = [oriC.middle for oriC in oriCs]

        # Step 6: Setting the last variables to the proper values
        for i in range(len(oriCs)):
            oriCs[i].z_score = Z_scores[i]
            oriCs[i].g_score = G_scores[i]
            oriCs[i].d_score = D_scores[i]
            oriCs[i].decision = decisions[i]

        self.oriCs = oriCs
        self.oriC_middles = oriC_middles
        self.Z_scores = Z_scores
        self.G_scores = G_scores
        self.D_scores = D_scores
        self.predictions = decisions
        self._set_best_pot_oriC_idx()

        if show_info:
            self.pretty_print_results()
        if plot_path is not None:
            self.plot_oriC_curves(plot_path)



"""Module for plotting disparity curves."""

from typing import Optional
import numpy as np
import matplotlib.pyplot as plt

import matplotlib
matplotlib.use("Agg")


def plot_Z_curve_3D(path: str, x: np.ndarray, y: np.ndarray, z: np.ndarray, name: Optional[str] = None) -> None:
    """
    3D-plot function with name as title. `x`, `y`, and `z` should be the three components of the Z-curve.

    -----------------------------------------------------------------------
    Example
    >>> from orcapy import Plotter, ORCA
    >>> import matplotlib.pyplot as plt
    >>> orca = ORCA.from_accession('NC_000913', 'example@email.com')
    >>> orca.find_oriCs()
    >>> Plotter.plot_curves_3D("path/to/plot.png", orca.x, orca.y, orca.z, name=orca.accession)
    """
    fig = plt.figure(figsize=(7, 7))
    ax = plt.axes(projection='3d')
    ax.set_xlabel('Purine vs. Pyrimidine', fontsize=10, labelpad=15)
    ax.set_ylabel('Amino vs. Keto', fontsize=10, labelpad=15)
    ax.set_zlabel('Weak vs. Strong H-bond', fontsize=10, labelpad=15)

    ax.plot3D(x, y, z, c='b', linewidth=0.8)
    if name is not None:
        ax.set_title(f'Z-Curve: {name}', fontsize=10, loc='center', pad=20)
    plt.savefig(path, dpi=500)


def plot_curves(path: str, curves: tuple[np.ndarray], labels: list[str], peaks: Optional[list[int]], name: Optional[str] = None) -> None:
    """
    Plots up to 6 different y-axes onto a single figure. Ideal for displaying multiple disparity curves in a single plot.
    If displaying more than 3 diferent axes at once, some manual adjustment of the subplot paramenters might be needed.
    - `path`   : string path to save location.
    - `curves` : list of lists with y-axis values.
    - `labels` : list of names of each curve in `curves`.
    - `peaks`  : optional, list with indices to plot onto the `curves`.
    - `name`   : optional, used in plot title.

    -----------------------------------------------------------------------
    Example
    >>> from orcapy import Plotter, ORCA
    >>> import matplotlib.pyplot as plt
    >>> orca = ORCA.from_accession('NC_000913', 'example@email.com')
    >>> orca.find_oriCs()
    >>> Plotter.plot_curves(path="path/to/plot.png", curves=[orca.x, orca.y, orca.gc], labels=['$x_n$', '$y_n$', '$GC_n$'], peaks=orca.oriC_middles, name=orca.accession)

    Alternatively call:

    >>> from orcapy import ORCA
    >>> import matplotlib.pyplot as plt
    >>> orca = ORCA.from_accession('NC_000913', 'example@email.com')
    >>> orca.find_oriCs()
    >>> orca.plot_oriC_curves(plot_path="path/to/plot.png")
    """
    x_len = str(len(curves[0]))
    if int(x_len[1]) <= 4:
        x_max = x_len[0] + str(int(x_len[1])+(5-int(x_len[1]))) + '0'*len(x_len[2:])
    else:
        x_max = str(int(x_len[0])+1) + '0'*len(x_len[1:])
    thing = max(int(x_max)//1000, 1)
    xthing = thing * 100
    ything = thing*2

    # Colourblind friendly list of colours, I think.
    color_list = ['#e31a1c', '#1f77b4', '#33a02c', '#6a3d9a', '#ff7f00', '#b15928']

    fig = plt.figure(figsize=(8.5, 4))
    fig.subplots_adjust(right=0.85 - (0.1 * max(len(curves)-2, 0)), bottom=0.25)
    base_ax = plt.axes()
    ax_list = [base_ax] + [base_ax.twinx() for i in range(len(curves) - 1)]

    offset = 1
    for axis in ax_list[1:]:
        axis.spines.right.set_position(("axes", offset))
        axis.yaxis.get_offset_text().set_x(offset)
        offset += 0.2

    good = False
    while not good:
        yticks_len = 0
        for i, ax in enumerate(curves):
            ubound, lbound = ything * round(max(ax)/ything), ything * round(min(ax)/ything)
            upper = ubound if max(ax) <= ubound else ubound + ything
            lower = lbound if min(ax) >= lbound else lbound - ything
            if len([x for x in range(lower, upper+ything, ything)]) > yticks_len:
                yticks_len = len([x for x in range(lower, upper+ything, ything)])
        if yticks_len < 6:
            yticks_len *= 2
            ything = ything // 2
        else:
            good = True
            break

    handle_list = []
    for i, ax in enumerate(curves):
        peaks_y = np.asarray([ax[j] for j in peaks])  # y refers to the y-axis coordinates, not the y-curve
        ax_list[i].plot(range(len(ax)), ax, color=color_list[i], zorder=2, label=labels[i])
        ax_list[i].scatter(peaks, peaks_y, marker='o', c='k', zorder=3, label='$\it{oriC}$')
        ax_list[i].tick_params(axis='y', colors=color_list[i])
        ax_list[i].ticklabel_format(axis='y', style='sci', scilimits=(3, 3), useMathText=True)

        lbound = ything * round(min(ax)/ything)
        lower = lbound if min(ax) >= lbound else lbound - ything
        yticks = [lower + ything*j for j in range(yticks_len)]
        ax_list[i].set_yticks(yticks)
        ax_list[i].set_ylim(min(yticks), max(yticks))

        h, _ = ax_list[i].get_legend_handles_labels()
        handle_list.extend(h[:-1])
        oriC_handle = h[-1]
    handle_list.append(oriC_handle)

    if name is not None:
        base_ax.set_title(f'2D Z-Curve: {name}', fontsize=10, loc='center', pad=20)

    ubound = xthing * round(len(ax)/xthing)
    upper = ubound if len(ax) <= ubound else ubound + xthing
    xticks = [x for x in range(0, upper+xthing, xthing)]
    base_ax.set_xticks(xticks)

    base_ax.ticklabel_format(axis='x', style='sci', scilimits=(3, 3), useMathText=True)
    base_ax.set_xlabel('Sequence length (bp)')
    base_ax.set_xlim(min(xticks), max(xticks))
    base_ax.grid(True, which='major', zorder=1)

    if len(peaks) != 0:
        l = labels + ['Prediction']
        n = len(curves)+1
    else:
        l = labels
        n = len(curves)

    plt.legend(
        handles=handle_list,
        labels=l,
        bbox_to_anchor=(0.12, -0.35, 0.75, .102),
        loc='center',
        ncol=n,
        mode="expand",
        borderaxespad=0.
    )
    plt.savefig(path, dpi=500)


def plot_skew(path: str, skew_array: np.ndarray, peaks: list[int], name: str) -> None:
    """Plots single skew diagram and its peaks"""

    fig = plt.figure()
    ax1 = plt.axes()

    peaks_y = np.asarray([skew_array[i] for i in peaks])

    ax1.set_title(f'GC-skew: {name}', fontsize=10, loc='center', pad=20)
    ax1.plot(range(len(skew_array)), skew_array, 'r', zorder=1)
    ax1.scatter(peaks, peaks_y, marker='X', c='k', zorder=2)
    plt.savefig(path, dpi=500)



import unittest as ut
import unittest.mock as m

from context import *


class TestORCA(ut.TestCase):
    def setUp(self):
        self.seq = "AGCATGCTAGTAGCTCGTCGTAGTACGATGCATGTCAGCTGATCGTACGATCGATAGATGACTACATGCACCACACGTACTACGATGATGATAGCTCA"
        self.gene_locs = [('dnaA', 0, 10), ('dnaN', 20, 30)]
        # remove later


    def test___make_1(self):
        ...

    
    def test_from_string_1(self):
        orca = ORCA.from_string(sequence=self.seq, gene_locations=self.gene_locs, accession="Test")

        exp_gene_locs = [("dnaA", Peak.from_calc_middle(0, 10, len(self.seq), 0)), ("dnaN", Peak.from_calc_middle(20, 30, len(self.seq), 0))]

        self.assertIsInstance(orca, ORCA)
        self.assertEqual(orca.accession, "Test")
        self.assertEqual(orca.seq, self.seq)
        self.assertEqual(orca.gene_locations, exp_gene_locs)
        self.assertEqual(orca.seq_len, len(self.seq))
        self.assertEqual(orca.version, 0)


    def test_from_string_2(self):
        orca = ORCA.from_string(sequence=self.seq)

        self.assertIsInstance(orca, ORCA)
        self.assertEqual(orca.accession, "Custom")
        self.assertEqual(orca.seq, self.seq)
        self.assertEqual(orca.gene_locations, [])
        self.assertEqual(orca.seq_len, len(self.seq))
        self.assertEqual(orca.version, 0)


    def test_from_gbk(self):
        ...


    def test_from_pkl(self):
        ...


    def test_from_accession(self):
        ...


    def test_setup_variables(self):
        ...


    def test_get_dnaa_boxes_with_mismatches(self):
        ...


    def test_calculate_disparity_curves(self):
        ...


    def test_analyse_disparity_curves(self):
        ...


    def test_calculate_Z_scores(self):
        ...


    def test_calculate_D_scores(self):
        ...


    def test_calculate_G_scores(self):
        ...


    def find_oriCs(self):
        ...


    def test_check_run_properly_1(self):
        orca = ORCA()
        with self.assertRaises(AttributeError) as context:
            orca.find_oriCs()

        exp_msg = "'ORCA' object has no attribute 'accession'"
        exp_note = "This object has not been instantiated properly."

        self.assertRegex(exp_msg, str(context.exception))
        self.assertRegex(exp_note, context.exception.__notes__[0])



    def test_check_run_properly_2(self):
        orca = ORCA.from_string(self.seq, gene_locations=self.gene_locs, accession="test")
        with self.assertRaises(AttributeError) as context:
            orca.plot_oriC_curves()

        exp_msg = "'ORCA' object has no attribute 'oriC_middles'"
        exp_note = "Cannot plot curves."

        self.assertRegex(exp_msg, str(context.exception))
        self.assertRegex(exp_note, context.exception.__notes__[0])



    def test_check_run_properly_3(self):
        orca = ORCA.from_string(self.seq, gene_locations=self.gene_locs, accession="test")
        with self.assertRaises(AttributeError) as context:
            orca.pretty_print_results()

        exp_msg = "'ORCA' object has no attribute 'oriC_middles'"
        exp_note = "Cannot print results."

        self.assertRegex(exp_msg, str(context.exception))
        self.assertRegex(exp_note, context.exception.__notes__[0])


if __name__ == "__main__":
    ut.main()


import os, io, pickle

import unittest as ut
import unittest.mock as m

from context import *

 
class TestBioFile(ut.TestCase):
    def setUp(self):

        self.acc = "NC_000913.3"
        self.email = "Some@email.address"
        self.inp = "Mock/folder/data/input"
        self.out = "Mock/folder/data/output"

        with open("data/input/NC_000913_3.pkl","rb") as fh:
            seq_rec = pickle.load(fh)
        self.seq_rec = seq_rec
        self.genes = ['dnaA', 'dnaN']

        dnaA = Peak.from_calc_middle(3882325, 3883729, 4641652, 0)
        dnaN = Peak.from_calc_middle(3881220, 3882321, 4641652, 0)
        oriC = Peak.from_calc_middle(3925743, 3925975, 4641652, 0)

        self.seq_dict = {
            'accession': 'NC_000913',
            'version': 3,
            'seq': seq_rec.seq,
            'gene_locations': [
                ('dnaN', dnaN),
                ('dnaA', dnaA)
            ],
            'NCBI_oriC': [('oriC', oriC)],
            'seq_len': 4641652
        }

        self.text_io = io.StringIO("""LOCUS       NC_000913            4641652 bp    DNA     circular CON 09-MAR-2022
                DEFINITION  Escherichia coli str. K-12 substr. MG1655, complete genome.
                ACCESSION   NC_000913
                VERSION     NC_000913.3

                gene            190..255
                                /gene="thrL"
                                /locus_tag="b0001"
                                /gene_synonym="ECK0001"
                                /db_xref="ASAP:ABE-0000006"
                                /db_xref="ECOCYC:EG11277"
                                /db_xref="GeneID:944742"
                CDS             190..255
                                /gene="thrL"
                                /locus_tag="b0001"
                                /gene_synonym="ECK0001"
                                /codon_start=1
                                /transl_table=11
                                /product="thr operon leader peptide"
                                /protein_id="NP_414542.1"
                                /db_xref="UniProtKB/Swiss-Prot:P0AD86"
                                /db_xref="ASAP:ABE-0000006"
                                /db_xref="ECOCYC:EG11277"
                                /db_xref="GeneID:944742"
                                /translation="MKRISTTITTTITITTGNGAG"
                """)


    def test_fetch_file_1_good(self):
        with m.patch('Bio.Entrez.efetch') as mock_efetch:
            mock_efetch.return_value = self.text_io
            x = BioFile.fetch_file(self.acc, self.email, "gbwithparts", None)
            self.assertEqual(self.text_io, x)


    def test_fetch_file_2_bad_request_with_api_key(self):
        with m.patch('Bio.Entrez.efetch') as mock_efetch:
            url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&rettype=&retmode=text&tool=biopython&email=Some%40email.address"
            code = 400; msg = "Bad Request"; hdrs = "Some headers"; fp = ""
            mock_efetch.side_effect = HTTPError(url, code, msg, hdrs, fp)

            api_message_1 = f' with API_key \'123\''
            api_message_2 = ' and if your API_key is correctly typed and linked to the given email'
            note = f'Unable to fetch accession: \'ACCESSION\' using \'Some@email.address\'{api_message_1}. Please check if the accession is of an existing chromosomal sequence{api_message_2}.'
            message = "HTTP Error 400: Bad Request"

            try:
                BioFile.fetch_file("ACCESSION", self.email, "gbwithparts", "123")
            except HTTPError as e:
                self.assertRegex(str(e), message)
                self.assertRegex(e.__notes__[0], note)


    def test_fetch_file_3_bad_request_no_api_key(self):
        with m.patch('Bio.Entrez.efetch') as mock_efetch:
            url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&rettype=&retmode=text&tool=biopython&email=Some%40email.address"
            code = 400; msg = "Bad Request"; hdrs = "Some headers"; fp = ""
            mock_efetch.side_effect = HTTPError(url, code, msg, hdrs, fp)

            note = f'Unable to fetch accession: \'ACCESSION\' using \'Some@email.address\'. Please check if the accession is of an existing chromosomal sequence.'
            message = "HTTP Error 400: Bad Request"
    
            try:
                BioFile.fetch_file("ACCESSION", self.email, "gbwithparts", None)
            except HTTPError as e:
                self.assertRegex(str(e), message)
                self.assertRegex(e.__notes__[0], note)


    def test_fetch_file_4_no_connetion(self):
        with m.patch('Bio.Entrez.efetch') as mock_efetch:
            mock_efetch.side_effect = URLError("My reason")
            note = 'You are fetching a file from the NCBI servers. Please make sure you have an internet connection to do so.'

            try:
                BioFile.fetch_file("ACCESSION", self.email, "gbwithparts", "123")
            except URLError as e:
                self.assertRegex(str(e), "My reason")
                self.assertRegex(e.__notes__[0], note)


    def test_parse_SeqRecord_1_good(self):
        res = BioFile.parse_SeqRecord(self.seq_rec, self.genes)
        self.assertEqual(res, self.seq_dict)


    def test_parse_SeqRecord_2_good(self):
        res = BioFile.parse_SeqRecord(self.seq_rec, ['dnaA'])
        self.seq_dict['gene_locations'].pop(0)
        self.assertEqual(res, self.seq_dict)


    def test_parse_SeqRecord_3_good(self):
        res = BioFile.parse_SeqRecord(self.seq_rec, [])
        self.seq_dict['gene_locations'] = []
        self.assertEqual(res, self.seq_dict)


    def test_save_gbk_1_good(self):
        # Set up the mock file object
        with m.patch('builtins.open', m.mock_open()) as mock_open, \
            m.patch('Bio.Entrez.efetch') as mock_efetch, \
            m.patch('os.listdir') as mock_listdir:

            # efetch return value gets closed by save_gbk, so get the value before the call
            to_write_a = """LOCUS       NC_000913            4641652 bp    DNA     circular CON 09-MAR-2022
                DEFINITION  Escherichia coli str. K-12 substr. MG1655, complete genome.
                ACCESSION   NC_000913
                VERSION     NC_000913.3
            """

            to_write_b = """
                gene            190..255
                                /gene="thrL"
                                /locus_tag="b0001"
                                /gene_synonym="ECK0001"
                                /db_xref="ASAP:ABE-0000006"
                                /db_xref="ECOCYC:EG11277"
                                /db_xref="GeneID:944742"
                CDS             190..255
                                /gene="thrL"
                                /locus_tag="b0001"
                                /gene_synonym="ECK0001"
                                /codon_start=1
                                /transl_table=11
                                /product="thr operon leader peptide"
                                /protein_id="NP_414542.1"
                                /db_xref="UniProtKB/Swiss-Prot:P0AD86"
                                /db_xref="ASAP:ABE-0000006"
                                /db_xref="ECOCYC:EG11277"
                                /db_xref="GeneID:944742"
                                /translation="MKRISTTITTTITITTGNGAG"
                """
            mock_efetch.return_value = self.text_io
            mock_listdir.return_value = ["NC_000000_1.gbk", "NC_000001_1.gbk", "NC_000002_1.gbk"]

            # Call the function
            BioFile.save_gbk(self.acc, self.email, self.out)

            # Assert that the file was opened and written to correctly
            path = os.path.join(self.out, 'NC_000913_3.gbk')
            mock_open.assert_called_with(path, 'w')
            mock_open().write.has_calls([to_write_a, to_write_b])


    def test_save_gbk_2_file_already_exists(self):
        # Set up the mock file object
        with m.patch('builtins.open', m.mock_open()) as mock_open, \
            m.patch('Bio.Entrez.efetch') as mock_efetch, \
            m.patch('os.path.exists') as mock_exists:

            mock_efetch.return_value = self.text_io
            mock_exists.return_value = True

            # Call the function
            message = f'\'NC_000913_3.gbk\' already exists in: {self.inp}'
            with self.assertRaisesRegex(FileExistsError, message):
                BioFile.save_gbk(self.acc, self.email, self.inp)


    def test_save_pkl_1_good(self):
        # Set up the mock file object
        with m.patch('builtins.open', m.mock_open()) as mock_open, \
            m.patch('Bio.Entrez.efetch') as mock_efetch, \
            m.patch('os.listdir') as mock_listdir, \
            m.patch('Bio.SeqIO.read') as mock_seqIO_read, \
            m.patch('pickle.dump') as mock_pickle_dump:

            mock_efetch.return_value = self.text_io
            mock_listdir.return_value = ["NC_000000_1.pkl", "NC_000001_1.pkl", "NC_000002_1.pkl"]
            mock_seqIO_read.return_value = self.seq_rec

            # Call the function
            BioFile.save_pkl(self.acc, self.email, self.out)

            # Assert that the file was opened and written to correctly
            path = os.path.join(self.out, 'NC_000913_3.pkl')
            mock_open.assert_called_once_with(path, 'wb')
            mock_pickle_dump.assert_called_once_with(self.seq_rec, mock_open.return_value)


    def test_save_pkl_2_file_already_exists(self):
        class MockSeqIO:
            pass

        # Set up the mock file object
        with m.patch('builtins.open', m.mock_open()) as mock_open, \
            m.patch('Bio.Entrez.efetch') as mock_efetch, \
            m.patch('os.path.exists') as mock_exists, \
            m.patch('Bio.SeqIO.read') as mock_read:
    
            mock_efetch.return_value = self.text_io
            mock_exists.return_value = True
            seqio = MockSeqIO()
            seqio.id = "NC_000913.3"
            mock_read.return_value = seqio

            # Call the function
            message = f'\'NC_000913_3.pkl\' already exists in: {self.inp}'
            with self.assertRaisesRegex(FileExistsError, message):
                BioFile.save_pkl(self.acc, self.email, self.inp)

if __name__ == "__main__":
    ut.main()


"""Module for handing and processing disparity curves"""

from itertools import combinations

import numpy as np
from scipy.signal import find_peaks

from orcapy.Peak import Peak


def process_curve(curve: np.ndarray, mode: str, window_size: int) -> list[Peak]:
    '''Runs the given 1D-array (curve) through all processing functions for oriC identification. Returns its peaks'''
    init_peaks = [Peak(peak, curve.shape[0], window_size) for peak in detect_peaks(curve)]
    accepted_peaks = filter_peaks(curve, mode, init_peaks)
    peaks_to_merge = Peak.get_peaks_to_merge(accepted_peaks)

    single_peaks = [x for x in accepted_peaks if not any(x in y for y in peaks_to_merge)]
    merged_peaks = [Peak.from_calc_middle(to_merge[0], to_merge[1], curve.shape[0],
                                          window_size) for to_merge in peaks_to_merge]
    return single_peaks + merged_peaks


def curve_combinations(peaks_list: tuple[list["Peak"]], seq_len: int) -> list:
    '''Get every matched_peaks combination for x, y, and gc.'''
    oriC_locations_list = []
    for peaks_i, peaks_j in combinations(peaks_list, 2):
        matched_peaks = Peak.match_peaks(peaks_i, peaks_j)
        oriC_locations_list.append(
            [Peak.from_calc_middle(matches[0], matches[1], seq_len, peaks_list[0][0].window_size)
             for matches in matched_peaks]
        )
    return oriC_locations_list


def detect_peaks(curve: np.ndarray) -> np.ndarray:
    '''Calculates peaks of 1D-np.array and returns its indices.'''
    maxima, _ = find_peaks(curve, distance=curve.shape[0]//12)
    maxima = np.append(maxima, curve.argmax())
    minima, _ = find_peaks(np.negative(curve), distance=curve.shape[0]//12)
    minima = np.append(minima, curve.argmin())
    return np.unique(np.concatenate((maxima, minima), axis=0))


def filter_peaks(curve: np.ndarray, mode: str, peaks: list[Peak]) -> list[Peak]:
    '''
    Filters the given peaks based on the type of extreme it should be and the area around the peak in two steps.
    - Filter 1: Check if peaks are actually the extreme in their windows. This accounts for local extremes on the sides of slopes.
    - Filter 2: Check if any windows intersect the window of another peak (circular DNA) and reject the one that was less extreme (min or max).
    ### Input:
    - `curve`          : 1D-np.array
    - `peaks`          : list of Peaks of curve
    - `mode`           : 'max'|'min'. Which type of extreme do you want to find?
    ### Return:
    - `accepted_peaks` : list of peaks that passed both filters
    '''
    rejected_peaks = []
    rejected_peaks.extend(_filter_within_windows(curve, mode, peaks))
    rejected_peaks.extend(_filter_intersecting_windows(curve, mode, peaks))

    # Return list of peaks that passed both filters
    return list(set(peaks).difference(rejected_peaks))


def _filter_within_windows(curve: np.ndarray, mode: str, peaks: list[Peak]):
    """Filter 1: Check if peaks are actually the extreme in their windows"""
    rejected = []
    for peak in peaks:
        if peak.split:
            # Determine which side of 0 the peak.middle (m) is:
            #     _______0                             0_______
            #     5'   m    3'                      5'   m    3'
            # -----------|-----------  vs.  -----------|-----------
            comparator_win = curve[peak.five_side:] if (peak.middle > peak.five_side) else curve[:peak.three_side]
        else:
            comparator_win = curve[peak.five_side:peak.three_side]

        if mode == 'max' and comparator_win.max() > curve[peak.middle]:
            rejected.append(peak)
        elif mode == 'min' and comparator_win.min() < curve[peak.middle]:
            rejected.append(peak)
    return rejected


def _filter_intersecting_windows(curve, mode, peaks):
    rejected = []
    for peak_i, peak_j in combinations(peaks, 2):
        # Filter 1: Check if any windows intersecting the window of peak i
        if peak_i.intersecting_windows(peak_j):
            # Either peaks at the beginning and end of the DNA sequence or a peak found by sp.find_peaks() that is very close to the global min/max
            if mode == 'max' and curve[peak_i.middle] > curve[peak_j.middle]:
                rejected.append(peak_j)
            elif mode == 'min' and curve[peak_i.middle] < curve[peak_j.middle]:
                rejected.append(peak_j)
            else:
                rejected.append(peak_i)
    return rejected



from typing import Union
from itertools import combinations, product

import numpy as np


class Peak():
    def __init__(self, middle: int, seq_len: int, window_size: int) -> "Peak":
        '''
        Peak object for keeping track of points with a window (i.e. border/buffer) surrounding them.
        Expected behaviour is that `self.middle` is always in the middle position between `self.five_side` and `self.three_side`.
        - `five_side`  : 5' side of the window.
        - `three_side` : 3' side of the window.

        Alternate constructors include: `from_calc_middle` and `from_edges`.
        '''
        self.middle = middle
        self.seq_len = seq_len
        self.window_size = window_size
        self.split = False

        self.z_score: int = 0
        self.g_score: int = 0
        self.d_score: int = 0
        self.decision: bool = None

        five_side = self.middle - self.window_size // 2
        three_side = self.middle + self.window_size // 2

        if five_side < 0:
            self.split = True
            five_side += self.seq_len-1
        if three_side > self.seq_len-1:
            self.split = True
            three_side -= self.seq_len-1

        self.five_side = five_side
        self.three_side = three_side

    @classmethod
    def from_calc_middle(cls, five_side, three_side, seq_len, window_size) -> "Peak":
        '''Calculates the middle of five_side and three_side and uses that as the middle'''
        return cls(cls.calc_middle(five_side, three_side, seq_len), seq_len, window_size)

    @classmethod
    def from_edges(cls, five_side, three_side, seq_len) -> "Peak":
        '''
        Calculates the middle of five_side and three_side and uses that as the middle 
        AND uses these edges to determine the window_size.
        '''
        return cls(cls.calc_middle(five_side, three_side, seq_len), seq_len, cls.calc_dist_points(five_side, three_side, seq_len))

    @staticmethod
    def calc_dist_points(a: int, b: int, curve_size: int) -> int:
        '''Calculate the distance of a to b in bp on a circular chromosome.'''
        dist_1 = max(a, b) - min(a, b)
        dist_2 = min(a, b) + curve_size - max(a, b)
        return min(dist_1, dist_2)

    @staticmethod
    def get_peaks_to_merge(peaks: list["Peak"]) -> list[tuple["Peak", "Peak"]]:
        """
        Get the indices that give the same value in the curve and are in eachothers window.
        Input:
        - `peaks`          : list of `Peaks` of a curve
        Return:
        - `peaks_to_merge` : Nested list. Each sublist contains the two Peaks that have to be merged
        """
        peaks_to_merge = []
        for peak_i, peak_j in combinations(peaks, 2):
            if peak_i.contains_point(peak_j.middle):
                peaks_to_merge.append((peak_i, peak_j))
        return peaks_to_merge

    @staticmethod
    def match_peaks(peaks_x: list["Peak"], peaks_y: list["Peak"]) -> list[tuple["Peak", "Peak"]]:
        '''Nested list of peaks from x that line up with peaks from y.'''
        matched_peaks = []
        for peak_x, peak_y in product(peaks_x, peaks_y):
            if peak_x.intersecting_windows(peak_y):
                matched_peaks.append((peak_x, peak_y))
        return matched_peaks

    @staticmethod
    def get_adjacency_matrix(peaks_a: list[Union["Peak", int]], peaks_b: list[Union["Peak", int]] = None, seq_len: int = None) -> np.ndarray:
        """
        Gets adjacency matrix for given list of `Peak` or `int` objects.
        The matrix can be between a list and the elements of itself or between the elements of two lists.
        All elements in each list must be of the same type.
        Elements in `peaks_a` must be the same type as those in `peaks_b`.
        If elements are integers, `seq_len` must be provided.
        """

        # Error handling and variable initialisation
        are_integers = True if isinstance(peaks_a[0], int) else False
        if are_integers and seq_len is None:
            raise ValueError('Provided list of integers, but did not provide `seq_len`.')
        if not all(isinstance(x, type(peaks_a[0])) for x in peaks_a[1:]):
            raise ValueError('Not all elements in `peaks_a` are of the same type.')

        if peaks_b is not None:
            if not all(isinstance(x, type(peaks_b[0])) for x in peaks_b[1:]):
                raise ValueError('Not all elements in `peaks_b` are of the same type.')
            if not isinstance(peaks_b[0], type(peaks_a[0])):
                raise ValueError('Elements is `peaks_a` are not the same type as `peaks_b`.')

            adj_mat = np.zeros((len(peaks_a), len(peaks_b)))
            iterator = product(enumerate(peaks_a), enumerate(peaks_b))
        else:
            adj_mat = np.zeros((len(peaks_a), len(peaks_a)))
            iterator = combinations(enumerate(peaks_a), r=2)

        for (i_a, a), (i_b, b) in iterator:
            dist = Peak.calc_dist_points(a, b, seq_len) if are_integers else Peak.calc_dist_points(
                a.middle, b.middle, a.seq_len)
            adj_mat[i_a, i_b] = dist
            if peaks_b is None:
                adj_mat[i_b, i_a] = dist
        return adj_mat

    @staticmethod
    def get_connected_components_idx(vertices: list[int], adj_mat: np.ndarray, threshold: int) -> list:
        """
        Recursively find connected groups in an undirected graph. 

        Parameters:
        - `vertices`  : list of indexes of whatever you want to group
        - `adj_mat`   : adjacency matrix of len(vertices) x len(vertices).
        - `threshold` : maximum size of an edge before two vertices are considered connected.
        """

        def _DFS_recurse(idx: int, adj_mat: np.ndarray, visited: list[bool], connected_list: list[int], threshold: int):
            """Private: used by get_connected_groups_idx for recursion"""
            visited[idx] = True
            connected_list.append(idx)
            for i in range(len(visited)):
                if adj_mat[i][idx] < threshold and not visited[i]:
                    _, _, visited, connected_list, _ = _DFS_recurse(i, adj_mat, visited, connected_list, threshold)
            return idx, adj_mat, visited, connected_list, threshold

        visited = [False] * len(vertices)
        connected_groups_idx = []
        for i in range(len(vertices)):
            if not visited[i]:
                group = []
                _, _, visited, group, _ = _DFS_recurse(i, adj_mat, visited, group, threshold=threshold)
                connected_groups_idx.append(group)
        return connected_groups_idx

    @staticmethod
    def select_connected_groups(peaks: list["Peak"], adj_mat: np.ndarray, threshold: int) -> list[list["Peak"]]:
        """
        Gets the connected components in an undirected graph and limits the within-distance of clusters to < threshold*3.

        Parameters:
        - `peaks`     : list of Peak objects you want to group
        - `adj_mat`   : adjacency matrix for each Peak to each other Peak
        - `threshold` : maximum distance between peaks on the sequence for which they are considered connected.
        """
        cg_i = Peak.get_connected_components_idx([i for i in range(len(peaks))], adj_mat, threshold)
        groups_to_process = [(group, threshold) for group in cg_i]
        accepted_groups_idx = []
        flag = False
        while len(groups_to_process) != 0:
            curr_group, curr_threshold = groups_to_process.pop(0)
            # check within distance of group
            for i, j in combinations(curr_group, r=2):
                # NOTE: Tune the threshold!!!!! especially check for large and realistic sequences!
                if adj_mat[i][j] > threshold*1.5:
                    flag = True
                    break
            # got flagged as too wide.
            if flag and curr_threshold > 0.25*threshold:
                # get peaks that correspond to the indexes in the flagged group
                curr_peaks = [peaks[i] for i in curr_group]
                mat = Peak.get_adjacency_matrix(curr_peaks)
                new_groups = Peak.get_connected_components_idx(curr_peaks, mat, curr_threshold)
                # revert indexing to that of the peaks list
                new_groups = [[curr_group[i] for i in group] for group in new_groups]
                # add new groups for processing
                groups_to_process.extend([(new_group, 0.75*curr_threshold) for new_group in new_groups])
                flag = False
            else:
                accepted_groups_idx.append(curr_group)

        connected_groups_vals = [[peaks[i] for i in idx_group] for idx_group in accepted_groups_idx]
        return connected_groups_vals

    @staticmethod
    def calc_middle(a: Union[int, "Peak"], b: Union[int, "Peak"], curve_size: int = None) -> int:
        """
        Calculate the distance between Peak a and b on circular DNA.
        Returns a new Peak.middle (int) in the middle of a and b
        """

        seq_len = a.seq_len if curve_size is None and isinstance(a, Peak) else curve_size
        seq_len = b.seq_len if seq_len is None and isinstance(b, Peak) else seq_len

        a, b = int(a), int(b)

        if seq_len is None:
            raise ValueError("No curve size found in any parameters. This method is for circular DNA only.")

        dist_1 = max(a, b) - min(a, b)
        dist_2 = min(a, b) + seq_len-1 - max(a, b)

        if dist_2 < dist_1:
            merged_idx = min(a, b) - dist_2//2
            if merged_idx < 0:
                merged_idx = seq_len-1 + merged_idx
        else:
            merged_idx = min(a, b) + dist_1//2
        return merged_idx

    def intersecting_windows(self, other: "Peak") -> bool:
        '''T|F wether self's window intersects with other's window.'''
        if not isinstance(other, Peak):
            raise ValueError(f'other is a {type(other)}, must be a Peak object')
        x = Peak.calc_dist_points(self.middle, other.middle, self.seq_len)
        y = self.window_size//2 + other.window_size//2
        return x < y

    def contains_point(self, point: Union["Peak", int, float]) -> bool:
        '''T|F wether a point is within a Peak's window'''
        p = point.middle if isinstance(point, Peak) else point
        return self.window_size // 2 >= Peak.calc_dist_points(self.middle, p, self.seq_len)

    def calc_dist(self, point_b: Union["Peak", int]):
        """Calculate the distance between self and another point on a cicular chromosome."""
        b = point_b.middle if isinstance(point_b, Peak) else point_b
        return Peak.calc_dist_points(self.middle, b, self.seq_len)

    def __hash__(self) -> int:
        '''very simple hash function. Does not include scores'''
        return hash(
            (self.middle, self.seq_len, self.window_size, self.split,
             self.five_side, self.three_side)
        )

    def __eq__(self, other: "Peak") -> bool:
        '''Simple equality check. Is not used in __lt__ and __gt__'''
        if not isinstance(other, Peak):
            return False
        return self.middle == other.middle \
            and self.seq_len == other.seq_len \
            and self.window_size == other.window_size \
            and self.split == other.split \
            and self.five_side == other.five_side \
            and self.three_side == other.three_side

    def __contains__(self, point: Union["Peak", int, float]) -> bool:
        """Return point in self's window. See: contains_point"""
        return self.contains_point(point)

    def __repr__(self) -> str:
        """Return repr(self)"""
        return f'Peak(middle={self.middle}, window_size={self.window_size}, seq_len={self.seq_len})'

    def __str__(self) -> str:
        """Return str(self)"""
        return str(self.middle)

    def __neg__(self) -> "Peak":
        """Returns -self with a Peak with negation of self.middle"""
        return Peak(-self.middle, self.seq_len, self.window_size)

    def __lt__(self, other: "Peak") -> bool:
        """Return self < other"""
        return self.middle < other.middle

    def __gt__(self, other: "Peak") -> bool:
        """Return self > other"""
        return self.middle > other.middle

    def __add__(self, other: Union["Peak", int, float]) -> Union["Peak", int, float]:
        """Return self + other"""
        if isinstance(other, Peak):
            return Peak(self.middle + other.middle, self.seq_len, self.window_size)
        elif isinstance(other, int) or isinstance(other, float):
            return self.middle + other

    def __sub__(self, other: Union["Peak", int, float]) -> Union["Peak", int, float]:
        """Return self - other"""
        if isinstance(other, Peak):
            return Peak(self.middle - other.middle, self.seq_len, self.window_size)
        elif isinstance(other, int) or isinstance(other, float):
            return self.middle - other

    def __radd__(self, other: Union["Peak", int, float]) -> Union["Peak", int, float]:
        """Return other + self"""
        return self.__add__(other)

    def __rsub__(self, other: Union["Peak", int, float]) -> Union["Peak", int, float]:
        """Return other - self"""
        return -self + other

    def __int__(self):
        """Return int of self.middle"""
        return int(self.middle)

    def __float__(self):
        """Return float of self.middle"""
        return float(self.middle)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="orcapy",
    version="1.0.1",
    description="Tool for predicting the origin of replication on circular bacterial chromosomes",
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=["test", "Machine_learning", "data"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ZoyavanMeel/ORCA",
    author="Zoya van Meel",
    author_email="zoyavanmeel@gmail.com",
    license="GPL-3.0",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy >= 1.22.3",
        "scikit-learn >= 1.4.0",
        "biopython >= 1.79",
        "scipy >= 1.8.0",
        "matplotlib >= 3.5.2"
    ],
    extras_require={
        "dev": ["twine>=5.0.0"],
    },
    python_requires=">=3.9",
)
